# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import sys

# Add topdir to import path
topdir = os.path.realpath(os.path.join(os.path.dirname(__file__) + "/../.."))
sys.path.insert(0, topdir)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import logging
import pmb


def is_dtb(path):
    if not os.path.isfile(path):
        return False
    with open(path, 'rb') as f:
        # Check FDT magic identifier (0xd00dfeed)
        return f.read(4) == b'\xd0\x0d\xfe\xed'


def get_mtk_label(path):
    """ Read the label from the MediaTek header of the kernel or ramdisk inside
        an extracted boot.img.
        :param path: to either the kernel or ramdisk extracted from boot.img
        :returns: * None: file does not exist or does not have MediaTek header
                  * Label string (e.g. "ROOTFS", "KERNEL") """
    if not os.path.exists(path):
        return None

    with open(path, 'rb') as f:
        # Check Mediatek header (0x88168858)
        if not f.read(4) == b'\x88\x16\x88\x58':
            return None
        f.seek(8)
        label = f.read(32).decode("utf-8").rstrip('\0')

        if label == "RECOVERY":
            logging.warning(
                "WARNING: This boot.img has MediaTek headers. Since you passed a"
                " recovery image instead of a regular boot.img, we can't tell what"
                " the ramdisk signature label is supposed to be, so we assume that"
                " it's the most common value, ROOTFS. There is a chance that this"
                " is wrong and it may not boot; in that case, run bootimg_analyze"
                " again with a regular boot.img. If this *is* a regular boot.img,"
                " replace the value of deviceinfo_bootimg_mtk_label_ramdisk with"
                " 'RECOVERY'.")
            return "ROOTFS"
        else:
            return label


def bootimg(args, path):
    if not os.path.exists(path):
        raise RuntimeError("Could not find file '" + path + "'")

    logging.info("NOTE: You will be prompted for your sudo/doas password, so"
                 " we can set up a chroot to extract and analyze your"
                 " boot.img file")
    pmb.chroot.apk.install(args, ["file", "unpackbootimg"])

    temp_path = pmb.chroot.other.tempfolder(args, "/tmp/bootimg_parser")
    bootimg_path = f"{args.work}/chroot_native{temp_path}/boot.img"

    # Copy the boot.img into the chroot temporary folder
    # and make it world readable
    pmb.helpers.run.root(args, ["cp", path, bootimg_path])
    pmb.helpers.run.root(args, ["chmod", "a+r", bootimg_path])

    file_output = pmb.chroot.user(args, ["file", "-b", "boot.img"],
                                  working_dir=temp_path,
                                  output_return=True).rstrip()
    if "android bootimg" not in file_output.lower():
        if "force" in args and args.force:
            logging.warning("WARNING: boot.img file seems to be invalid, but"
                            " proceeding anyway (-f specified)")
        else:
            logging.info("NOTE: If you are sure that your file is a valid"
                         " boot.img file, you could force the analysis"
                         " with: 'pmbootstrap bootimg_analyze " + path +
                         " -f'")
            if ("linux kernel" in file_output.lower() or
                    "ARM OpenFirmware FORTH Dictionary" in file_output):
                raise RuntimeError("File is a Kernel image, you might need the"
                                   " 'heimdall-isorec' flash method. See also:"
                                   " <https://wiki.postmarketos.org/wiki/"
                                   "Deviceinfo_flash_methods>")
            else:
                raise RuntimeError("File is not an Android boot.img. (" +
                                   file_output + ")")

    # Extract all the files
    pmb.chroot.user(args, ["unpackbootimg", "-i", "boot.img"],
                    working_dir=temp_path)

    output = {}
    header_version = 0
    # Get base, offsets, pagesize, cmdline and qcdt info
    # This file does not exist for example for qcdt images
    if os.path.isfile(f"{bootimg_path}-header_version"):
        with open(f"{bootimg_path}-header_version", 'r') as f:
            header_version = int(f.read().replace('\n', ''))
            output["header_version"] = str(header_version)

    if header_version >= 3:
        output["pagesize"] = "4096"
    else:
        with open(f"{bootimg_path}-base", 'r') as f:
            output["base"] = ("0x%08x" % int(f.read().replace('\n', ''), 16))
        with open(f"{bootimg_path}-kernel_offset", 'r') as f:
            output["kernel_offset"] = ("0x%08x"
                                       % int(f.read().replace('\n', ''), 16))
        with open(f"{bootimg_path}-ramdisk_offset", 'r') as f:
            output["ramdisk_offset"] = ("0x%08x"
                                        % int(f.read().replace('\n', ''), 16))
        with open(f"{bootimg_path}-second_offset", 'r') as f:
            output["second_offset"] = ("0x%08x"
                                       % int(f.read().replace('\n', ''), 16))
        with open(f"{bootimg_path}-tags_offset", 'r') as f:
            output["tags_offset"] = ("0x%08x"
                                     % int(f.read().replace('\n', ''), 16))
        with open(f"{bootimg_path}-pagesize", 'r') as f:
            output["pagesize"] = f.read().replace('\n', '')

        if header_version == 2:
            with open(f"{bootimg_path}-dtb_offset", 'r') as f:
                output["dtb_offset"] = ("0x%08x"
                                        % int(f.read().replace('\n', ''), 16))

    if get_mtk_label(f"{bootimg_path}-kernel") is not None:
        output["mtk_label_kernel"] = get_mtk_label(f"{bootimg_path}-kernel")
    if get_mtk_label(f"{bootimg_path}-ramdisk") is not None:
        output["mtk_label_ramdisk"] = get_mtk_label(f"{bootimg_path}-ramdisk")

    output["qcdt"] = ("true" if os.path.isfile(f"{bootimg_path}-dt") and
                      os.path.getsize(f"{bootimg_path}-dt") > 0 else "false")

    output["dtb_second"] = ("true" if is_dtb(f"{bootimg_path}-second")
                            else "false")

    with open(f"{bootimg_path}-cmdline", 'r') as f:
        output["cmdline"] = f.read().replace('\n', '')

    # Cleanup
    pmb.chroot.root(args, ["rm", "-r", temp_path])

    return output



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.helpers.run_core


def user(args, cmd, working_dir=None, output="log", output_return=False,
         check=None, env={}, sudo=False):
    """
    Run a command on the host system as user.

    :param env: dict of environment variables to be passed to the command, e.g.
                {"JOBS": "5"}

    See pmb.helpers.run_core.core() for a detailed description of all other
    arguments and the return value.
    """
    # Readable log message (without all the escaping)
    msg = "% "
    for key, value in env.items():
        msg += key + "=" + value + " "
    if working_dir:
        msg += "cd " + working_dir + "; "
    msg += " ".join(cmd)

    # Add environment variables and run
    env = env.copy()
    pmb.helpers.run_core.add_proxy_env_vars(env)
    if env:
        cmd = ["sh", "-c", pmb.helpers.run_core.flat_cmd(cmd, env=env)]
    return pmb.helpers.run_core.core(args, msg, cmd, working_dir, output,
                                     output_return, check, sudo)


def root(args, cmd, working_dir=None, output="log", output_return=False,
         check=None, env={}):
    """
    Run a command on the host system as root, with sudo or doas.

    :param env: dict of environment variables to be passed to the command, e.g.
                {"JOBS": "5"}

    See pmb.helpers.run_core.core() for a detailed description of all other
    arguments and the return value.
    """
    env = env.copy()
    pmb.helpers.run_core.add_proxy_env_vars(env)

    if env:
        cmd = ["sh", "-c", pmb.helpers.run_core.flat_cmd(cmd, env=env)]
    cmd = pmb.config.sudo(cmd)

    return user(args, cmd, working_dir, output, output_return, check, env,
                True)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
""" Common code for git tests """
import os

import pmb.helpers.git
import pmb.helpers.run
import shutil


def prepare_tmpdir(args, monkeypatch, tmpdir):
    """ Prepare git repositories in tmpdir, and override related functions.

        Git repositories:
        * local: like local clone of pmaports.git
        * remote: emulate a remote repository that we can add to "local", so
                  we can pass the tracking-remote tests in pmb.helpers.git.pull
        * remote2: unexpected remote that pmbootstrap can complain about

        Function overrides:
        * pmb.helpers.git.get_path: always return path to "local" repo
        * pmb.helpers.git.get_upstream_remote: always return "origin"

        :returns: path_local, run_git
                  * path_local: path to "local" repo
                  * run_git(git_args, repo="local"): convenience function """
    # Directory structure
    tmpdir = str(tmpdir)
    path_local = tmpdir + "/local"
    path_remote = tmpdir + "/remote"
    path_remote2 = tmpdir + "/remote2"
    os.makedirs(path_local)
    os.makedirs(path_remote)
    os.makedirs(path_remote2)

    def run_git(git_args, repo="local"):
        path = tmpdir + "/" + repo
        pmb.helpers.run.user(args, ["git"] + git_args, path, "stdout", output_return=True)

    # Remote repos
    run_git(["init", "-b", "master", "."], "remote")
    run_git(["commit", "--allow-empty", "-m", "commit: remote"], "remote")
    run_git(["init", "-b", "master", "."], "remote2")
    run_git(["commit", "--allow-empty", "-m", "commit: remote2"], "remote2")

    # Local repo (with master -> origin2/master)
    run_git(["init", "-b", "master", "."])
    run_git(["remote", "add", "-f", "origin", path_remote])
    run_git(["remote", "add", "-f", "origin2", path_remote2])
    run_git(["checkout", "-b", "master", "--track", "origin2/master"])

    # Override get_path()
    def get_path(args, name_repo):
        return path_local
    monkeypatch.setattr(pmb.helpers.git, "get_path", get_path)

    # Override get_upstream_remote()
    def get_u_r(args, name_repo):
        return "origin"
    monkeypatch.setattr(pmb.helpers.git, "get_upstream_remote", get_u_r)

    return path_local, run_git

def copy_dotgit(args, tmpdir):
    shutil.copytree(args.aports + "/.git", tmpdir + "/.git", ignore_dangling_symlinks=True)



#!/usr/bin/env python3
# -*- encoding: UTF-8 -*-
# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
# PYTHON_ARGCOMPLETE_OK
import sys
import pmb

# A convenience wrapper for running pmbootstrap from the git repository. This
# script is not part of the python packaging, so don't add more logic here!
if __name__ == "__main__":
    sys.exit(pmb.main())



# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
""" Test preserving HTTP_PROXY and other proxy env vars with all pmbootstrap
    run functions. """
import os
import pytest
import sys

import pmb_test  # noqa
import pmb.chroot.root
import pmb.chroot.user
import pmb.helpers.run
import pmb.helpers.run_core


@pytest.fixture
def args(request):
    import pmb.parse
    sys.argv = ["pmbootstrap.py", "chroot"]
    args = pmb.parse.arguments()
    args.log = args.work + "/log_testsuite.txt"
    pmb.helpers.logging.init(args)
    request.addfinalizer(pmb.helpers.logging.logfd.close)
    return args


def test_proxy_user(args, monkeypatch):
    func = pmb.helpers.run.user
    monkeypatch.setattr(os, "environ", {"HTTP_PROXY": "testproxy"})
    ret = func(args, ["sh", "-c", 'echo "$HTTP_PROXY"'], output_return=True)
    assert ret == "testproxy\n"


def test_proxy_root(args, monkeypatch):
    func = pmb.helpers.run.root
    monkeypatch.setattr(os, "environ", {"HTTP_PROXY": "testproxy"})
    ret = func(args, ["sh", "-c", 'echo "$HTTP_PROXY"'], output_return=True)
    assert ret == "testproxy\n"


def test_proxy_chroot_user(args, monkeypatch):
    func = pmb.chroot.user
    monkeypatch.setattr(os, "environ", {"HTTP_PROXY": "testproxy"})
    ret = func(args, ["sh", "-c", 'echo "$HTTP_PROXY"'], output_return=True)
    assert ret == "testproxy\n"


def test_proxy_chroot_root(args, monkeypatch):
    func = pmb.chroot.root
    monkeypatch.setattr(os, "environ", {"HTTP_PROXY": "testproxy"})
    ret = func(args, ["sh", "-c", 'echo "$HTTP_PROXY"'], output_return=True)
    assert ret == "testproxy\n"



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
""" Test pmb/chroot/mount.py """
import os
import pytest
import sys

import pmb_test  # noqa
import pmb.chroot


@pytest.fixture
def args(tmpdir, request):
    import pmb.parse
    sys.argv = ["pmbootstrap", "init"]
    args = pmb.parse.arguments()
    args.log = args.work + "/log_testsuite.txt"
    pmb.helpers.logging.init(args)
    request.addfinalizer(pmb.helpers.logging.logfd.close)
    return args


def test_chroot_mount(args):
    suffix = "native"
    mnt_dir = f"{args.work}/chroot_native/mnt/pmbootstrap"

    # Run something in the chroot to have the dirs created
    pmb.chroot.root(args, ["true"])
    assert os.path.exists(mnt_dir)
    assert os.path.exists(f"{mnt_dir}/packages")

    # Umount everything, like in pmb.install.install_system_image
    pmb.helpers.mount.umount_all(args, f"{args.work}/chroot_{suffix}")

    # Remove all /mnt/pmbootstrap dirs
    pmb.chroot.remove_mnt_pmbootstrap(args, suffix)
    assert not os.path.exists(mnt_dir)

    # Run again: it should not crash
    pmb.chroot.remove_mnt_pmbootstrap(args, suffix)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import fnmatch
import pytest
import sys

import pmb_test  # noqa
import pmb.build
import pmb.chroot.apk


@pytest.fixture
def args(tmpdir, request):
    import pmb.parse
    sys.argv = ["pmbootstrap.py", "init"]
    args = pmb.parse.arguments()
    args.log = args.work + "/log_testsuite.txt"
    pmb.helpers.logging.init(args)
    request.addfinalizer(pmb.helpers.logging.logfd.close)
    return args


def test_install_build(monkeypatch, args):
    func = pmb.chroot.apk.install_build
    ret_apkindex_package = None

    def fake_build_package(args, package, arch):
        return "build-pkg"
    monkeypatch.setattr(pmb.build, "package", fake_build_package)

    def fake_apkindex_package(args, package, arch, must_exist):
        return ret_apkindex_package
    monkeypatch.setattr(pmb.parse.apkindex, "package", fake_apkindex_package)

    package = "hello-world"
    arch = "x86_64"

    # invoked as pmb install, build_pkgs_on_install disabled
    args.action = "install"
    args.build_pkgs_on_install = False
    with pytest.raises(RuntimeError) as e:
        func(args, package, arch)
    assert "no binary package found" in str(e.value)

    # invoked as pmb install, build_pkgs_on_install disabled, binary exists
    args.action = "install"
    args.build_pkgs_on_install = False
    ret_apkindex_package = {"pkgname": "hello-world"}
    assert func(args, package, arch) is None

    # invoked as pmb install, build_pkgs_on_install enabled
    args.action = "install"
    args.build_pkgs_on_install = True
    assert func(args, package, arch) == "build-pkg"

    # invoked as not pmb install
    args.action = "chroot"
    args.build_pkgs_on_install = False
    assert func(args, package, arch) == "build-pkg"


def test_packages_split_to_add_del():
    packages = ["hello", "!test", "hello2", "test2", "!test3"]

    to_add, to_del = pmb.chroot.apk.packages_split_to_add_del(packages)
    assert to_add == ["hello", "hello2", "test2"]
    assert to_del == ["test", "test3"]


def test_packages_get_locally_built_apks(monkeypatch, args):
    args.assume_yes = True

    arch = pmb.config.arch_native
    packages = ["hello-world",  # will exist in repo and locally
                "postmarketos-base",  # will exist in repo only
                "package-that-does-not-exist"]  # will not exist at all

    pmb.chroot.zap(args, pkgs_local=True)
    pmb.build.package(args, "hello-world", force=True)

    ret = pmb.chroot.apk.packages_get_locally_built_apks(args, packages, arch)
    assert len(ret) == 1
    assert fnmatch.fnmatch(ret[0], "*/hello-world-*.apk")


def test_install_run_apk(monkeypatch, args):
    global cmds_progress
    global cmds

    func = pmb.chroot.apk.install_run_apk
    suffix = "chroot_native"

    def fake_chroot_root(args, command, suffix):
        global cmds
        cmds += [command]
    monkeypatch.setattr(pmb.chroot, "root", fake_chroot_root)

    def fake_apk_progress(args, command, chroot, suffix):
        global cmds_progress
        cmds_progress += [command]
    monkeypatch.setattr(pmb.helpers.apk, "apk_with_progress", fake_apk_progress)

    def reset_cmds():
        global cmds_progress, cmds
        cmds = []
        cmds_progress = []

    # Simple add
    reset_cmds()
    to_add = ["postmarketos-base", "device-ppp"]
    to_add_local = []
    to_del = []
    func(args, to_add, to_add_local, to_del, suffix)
    assert cmds_progress == [["apk", "add", "postmarketos-base", "device-ppp",
                              "--no-interactive"]]
    assert cmds == []

    # Add and delete
    reset_cmds()
    to_add = ["postmarketos-base", "device-ppp"]
    to_add_local = []
    to_del = ["osk-sdl"]
    func(args, to_add, to_add_local, to_del, suffix)
    assert cmds_progress == [["apk", "add", "postmarketos-base", "device-ppp",
                              "--no-interactive"]]
    assert cmds == [["apk", "--no-progress", "del", "osk-sdl",
                     "--no-interactive"]]

    # Add with local package
    reset_cmds()
    to_add = ["postmarketos-base", "device-ppp"]
    to_add_local = ["/tmp/device-ppp.apk"]
    to_del = []
    func(args, to_add, to_add_local, to_del, suffix)
    assert cmds_progress == [["apk", "add", "postmarketos-base", "device-ppp",
                              "--no-interactive"]]
    assert cmds == [["apk", "--no-progress", "add", "-u", "--virtual",
                     ".pmbootstrap", "/tmp/device-ppp.apk", "--no-interactive"],
                    ["apk", "--no-progress", "del", ".pmbootstrap",
                     "--no-interactive"]]

    # Add with --no-network
    reset_cmds()
    args.offline = True
    to_add = ["hello-world"]
    to_add_local = []
    to_del = []
    func(args, to_add, to_add_local, to_del, suffix)
    assert cmds_progress == [["apk", "--no-network", "add", "hello-world",
                              "--no-interactive"]]
    assert cmds == []

    # Package name starting with '-'
    reset_cmds()
    to_add = ["hello-world", "--allow-untrusted"]
    to_add_local = []
    to_del = []
    with pytest.raises(ValueError) as e:
        func(args, to_add, to_add_local, to_del, suffix)
    assert "Invalid package name" in str(e.value)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import glob
import logging
import os
import pathlib

import pmb.build
import pmb.config
import pmb.chroot
import pmb.chroot.apk
import pmb.helpers.run
import pmb.parse.arch


def init_abuild_minimal(args, suffix="native"):
    """ Initialize a minimal chroot with abuild where one can do
        'abuild checksum'. """
    marker = f"{args.work}/chroot_{suffix}/tmp/pmb_chroot_abuild_init_done"
    if os.path.exists(marker):
        return

    pmb.chroot.apk.install(args, ["abuild"], suffix, build=False)

    # Fix permissions
    pmb.chroot.root(args, ["chown", "root:abuild",
                           "/var/cache/distfiles"], suffix)
    pmb.chroot.root(args, ["chmod", "g+w",
                           "/var/cache/distfiles"], suffix)

    # Add user to group abuild
    pmb.chroot.root(args, ["adduser", "pmos", "abuild"], suffix)

    pathlib.Path(marker).touch()


def init(args, suffix="native"):
    """ Initialize a chroot for building packages with abuild. """
    marker = f"{args.work}/chroot_{suffix}/tmp/pmb_chroot_build_init_done"
    if os.path.exists(marker):
        return

    init_abuild_minimal(args, suffix)

    # Initialize chroot, install packages
    pmb.chroot.apk.install(args, pmb.config.build_packages, suffix,
                           build=False)

    # Generate package signing keys
    chroot = args.work + "/chroot_" + suffix
    if not os.path.exists(args.work + "/config_abuild/abuild.conf"):
        logging.info("(" + suffix + ") generate abuild keys")
        pmb.chroot.user(args, ["abuild-keygen", "-n", "-q", "-a"],
                        suffix, env={"PACKAGER": "pmos <pmos@local>"})

        # Copy package signing key to /etc/apk/keys
        for key in glob.glob(chroot +
                             "/mnt/pmbootstrap/abuild-config/*.pub"):
            key = key[len(chroot):]
            pmb.chroot.root(args, ["cp", key, "/etc/apk/keys/"], suffix)

    apk_arch = pmb.parse.arch.from_chroot_suffix(args, suffix)

    # Add apk wrapper that runs native apk and lies about arch
    if pmb.parse.arch.cpu_emulation_required(apk_arch) and \
            not os.path.exists(chroot + "/usr/local/bin/abuild-apk"):
        with open(chroot + "/tmp/apk_wrapper.sh", "w") as handle:
            content = f"""
                #!/bin/sh
                export LD_PRELOAD_PATH=/native/usr/lib:/native/lib
                args=""
                for arg in "$@"; do
                    if [ "$arg" == "--print-arch" ]; then
                        echo "{apk_arch}"
                        exit 0
                    fi
                    args="$args $arg"
                done
                /native/usr/bin/abuild-apk $args
            """
            lines = content.split("\n")[1:]
            for i in range(len(lines)):
                lines[i] = lines[i][16:]
            handle.write("\n".join(lines))
        pmb.chroot.root(args, ["cp", "/tmp/apk_wrapper.sh",
                               "/usr/local/bin/abuild-apk"], suffix)
        pmb.chroot.root(args, ["chmod", "+x", "/usr/local/bin/abuild-apk"], suffix)

    # abuild.conf: Don't clean the build folder after building, so we can
    # inspect it afterwards for debugging
    pmb.chroot.root(args, ["sed", "-i", "-e", "s/^CLEANUP=.*/CLEANUP=''/",
                           "/etc/abuild.conf"], suffix)

    # abuild.conf: Don't clean up installed packages in strict mode, so
    # abuild exits directly when pressing ^C in pmbootstrap.
    pmb.chroot.root(args, ["sed", "-i", "-e",
                           "s/^ERROR_CLEANUP=.*/ERROR_CLEANUP=''/",
                           "/etc/abuild.conf"], suffix)

    pathlib.Path(marker).touch()


def init_compiler(args, depends, cross, arch):
    cross_pkgs = ["ccache-cross-symlinks", "abuild"]
    if "gcc4" in depends:
        cross_pkgs += ["gcc4-" + arch]
    elif "gcc6" in depends:
        cross_pkgs += ["gcc6-" + arch]
    else:
        cross_pkgs += ["gcc-" + arch, "g++-" + arch]
    if "clang" in depends or "clang-dev" in depends:
        cross_pkgs += ["clang"]
    if cross == "crossdirect":
        cross_pkgs += ["crossdirect"]
        if "rust" in depends or "cargo" in depends:
            if args.ccache:
                cross_pkgs += ["sccache"]
            # crossdirect for rust installs all build dependencies in the
            # native chroot too, as some of them can be required for building
            # native macros / build scripts
            cross_pkgs += depends

    pmb.chroot.apk.install(args, cross_pkgs)
